import keras
import numpy as np

from keras_hub.src.models.d_fine.d_fine_attention import DFineMultiheadAttention
from keras_hub.src.utils.keras_utils import clone_initializer


class DFineEncoderLayer(keras.layers.Layer):
    """Single encoder layer for D-FINE models.

    This layer is the fundamental building block of the `DFineEncoder`. It
    implements a standard transformer encoder layer with multi-head
    self-attention (`DFineMultiheadAttention`) and a feed-forward network. It is
    used to process and refine the feature sequences from the CNN backbone.

    Args:
        normalize_before: bool, Whether to apply layer normalization before
            the attention and feed-forward sub-layers (pre-norm) or after
            (post-norm).
        encoder_hidden_dim: int, Hidden dimension size of the encoder.
        num_attention_heads: int, Number of attention heads in multi-head
            attention.
        dropout: float, Dropout probability applied to attention outputs and
            feed-forward outputs.
        layer_norm_eps: float, Small constant added to the denominator for
            numerical stability in layer normalization.
        encoder_activation_function: str, Activation function used in the
            feed-forward network.
        activation_dropout: float, Dropout probability applied after the
            activation function in the feed-forward network.
        encoder_ffn_dim: int, Hidden dimension size of the feed-forward network.
        **kwargs: Additional keyword arguments passed to the parent class.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
    """

    def __init__(
        self,
        normalize_before,
        encoder_hidden_dim,
        num_attention_heads,
        dropout,
        layer_norm_eps,
        encoder_activation_function,
        activation_dropout,
        encoder_ffn_dim,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.normalize_before = normalize_before
        self.encoder_hidden_dim = encoder_hidden_dim
        self.num_attention_heads = num_attention_heads
        self.dropout_rate = dropout
        self.layer_norm_eps = layer_norm_eps
        self.encoder_activation_function = encoder_activation_function
        self.activation_dropout_rate = activation_dropout
        self.encoder_ffn_dim = encoder_ffn_dim
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.self_attn = DFineMultiheadAttention(
            embedding_dim=self.encoder_hidden_dim,
            num_heads=self.num_attention_heads,
            dropout=self.dropout_rate,
            dtype=self.dtype_policy,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            bias_initializer=clone_initializer(self.bias_initializer),
            name="self_attn",
        )
        self.self_attn_layer_norm = keras.layers.LayerNormalization(
            epsilon=self.layer_norm_eps,
            name="self_attn_layer_norm",
            dtype=self.dtype_policy,
        )
        self.dropout_layer = keras.layers.Dropout(
            rate=self.dropout_rate,
            name="dropout_layer",
            dtype=self.dtype_policy,
        )
        self.activation_fn_layer = keras.layers.Activation(
            self.encoder_activation_function,
            name="activation_fn_layer",
            dtype=self.dtype_policy,
        )
        self.activation_dropout_layer = keras.layers.Dropout(
            rate=self.activation_dropout_rate,
            name="activation_dropout_layer",
            dtype=self.dtype_policy,
        )
        self.fc1 = keras.layers.Dense(
            self.encoder_ffn_dim,
            name="fc1",
            dtype=self.dtype_policy,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            bias_initializer=clone_initializer(self.bias_initializer),
        )
        self.fc2 = keras.layers.Dense(
            self.encoder_hidden_dim,
            name="fc2",
            dtype=self.dtype_policy,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            bias_initializer=clone_initializer(self.bias_initializer),
        )
        self.final_layer_norm = keras.layers.LayerNormalization(
            epsilon=self.layer_norm_eps,
            name="final_layer_norm",
            dtype=self.dtype_policy,
        )

    def build(self, input_shape):
        self.self_attn.build(input_shape)
        self.self_attn_layer_norm.build(input_shape)
        self.fc1.build(input_shape)
        self.fc2.build((input_shape[0], input_shape[1], self.encoder_ffn_dim))
        self.final_layer_norm.build(input_shape)
        super().build(input_shape)

    def call(
        self,
        hidden_states,
        attention_mask=None,
        position_embeddings=None,
        output_attentions=False,
        training=None,
    ):
        residual = hidden_states
        if self.normalize_before:
            hidden_states = self.self_attn_layer_norm(
                hidden_states, training=training
            )
        hidden_states, attn_weights = self.self_attn(
            hidden_states=hidden_states,
            attention_mask=attention_mask,
            position_embeddings=position_embeddings,
            output_attentions=output_attentions,
            training=training,
        )
        hidden_states = self.dropout_layer(hidden_states, training=training)
        hidden_states = residual + hidden_states
        if not self.normalize_before:
            hidden_states = self.self_attn_layer_norm(
                hidden_states, training=training
            )
        if self.normalize_before:
            hidden_states = self.final_layer_norm(
                hidden_states, training=training
            )
        residual_ffn = hidden_states
        hidden_states = self.fc1(hidden_states)
        hidden_states = self.activation_fn_layer(hidden_states)
        hidden_states = self.activation_dropout_layer(
            hidden_states, training=training
        )
        hidden_states = self.fc2(hidden_states)
        hidden_states = self.dropout_layer(hidden_states, training=training)
        hidden_states = residual_ffn + hidden_states
        if not self.normalize_before:
            hidden_states = self.final_layer_norm(
                hidden_states, training=training
            )
        if training:
            dtype_name = keras.backend.standardize_dtype(self.compute_dtype)
            if dtype_name == "float16":
                clamp_value = np.finfo(np.float16).max - 1000.0
            else:  # float32, bfloat16
                clamp_value = np.finfo(np.float32).max - 1000.0
            hidden_states = keras.ops.clip(
                hidden_states, x_min=-clamp_value, x_max=clamp_value
            )
        if output_attentions:
            return hidden_states, attn_weights
        return hidden_states

    def compute_output_spec(
        self,
        hidden_states,
        attention_mask=None,
        position_embeddings=None,
        output_attentions=False,
        training=None,
    ):
        attn_output_spec = self.self_attn.compute_output_spec(
            hidden_states,
            position_embeddings,
            attention_mask,
            output_attentions,
        )
        if output_attentions:
            hidden_states_output_spec, self_attn_weights_spec = attn_output_spec
            return hidden_states_output_spec, self_attn_weights_spec
        return attn_output_spec

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "normalize_before": self.normalize_before,
                "encoder_hidden_dim": self.encoder_hidden_dim,
                "num_attention_heads": self.num_attention_heads,
                "dropout": self.dropout_rate,
                "layer_norm_eps": self.layer_norm_eps,
                "encoder_activation_function": self.encoder_activation_function,
                "activation_dropout": self.activation_dropout_rate,
                "encoder_ffn_dim": self.encoder_ffn_dim,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
            }
        )
        return config


class DFineEncoder(keras.layers.Layer):
    """Multi-layer encoder for D-FINE models.

    This layer implements a stack of `DFineEncoderLayer` instances. It is used
    within the `DFineHybridEncoder` to apply transformer-based processing to
    the feature maps from the CNN backbone, creating rich contextual
    representations before they are passed to the FPN/PAN pathways.

    Args:
        normalize_before: bool, Whether to apply layer normalization before
            the attention and feed-forward sub-layers (pre-norm) or after
            (post-norm) in each encoder layer.
        encoder_hidden_dim: int, Hidden dimension size of the encoder layers.
        num_attention_heads: int, Number of attention heads in multi-head
            attention for each layer.
        dropout: float, Dropout probability applied to attention outputs and
            feed-forward outputs in each layer.
        layer_norm_eps: float, Small constant added to the denominator for
            numerical stability in layer normalization.
        encoder_activation_function: str, Activation function used in the
            feed-forward networks of each layer.
        activation_dropout: float, Dropout probability applied after the
            activation function in the feed-forward networks.
        encoder_ffn_dim: int, Hidden dimension size of the feed-forward
            networks in each layer.
        num_encoder_layers: int, Number of encoder layers in the stack.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights of each layer. Defaults to
            `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights of each layer. Defaults to
            `"zeros"`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        normalize_before,
        encoder_hidden_dim,
        num_attention_heads,
        dropout,
        layer_norm_eps,
        encoder_activation_function,
        activation_dropout,
        encoder_ffn_dim,
        num_encoder_layers,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.normalize_before = normalize_before
        self.encoder_hidden_dim = encoder_hidden_dim
        self.num_attention_heads = num_attention_heads
        self.dropout_rate = dropout
        self.layer_norm_eps = layer_norm_eps
        self.encoder_activation_function = encoder_activation_function
        self.activation_dropout_rate = activation_dropout
        self.encoder_ffn_dim = encoder_ffn_dim
        self.num_encoder_layers = num_encoder_layers
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.encoder_layer = []
        for i in range(self.num_encoder_layers):
            layer = DFineEncoderLayer(
                normalize_before=self.normalize_before,
                encoder_hidden_dim=self.encoder_hidden_dim,
                num_attention_heads=self.num_attention_heads,
                dropout=self.dropout_rate,
                layer_norm_eps=self.layer_norm_eps,
                encoder_activation_function=self.encoder_activation_function,
                activation_dropout=self.activation_dropout_rate,
                encoder_ffn_dim=self.encoder_ffn_dim,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                dtype=self.dtype_policy,
                name=f"encoder_layer_{i}",
            )
            self.encoder_layer.append(layer)

    def build(self, input_shape):
        current_input_shape_for_layer = input_shape
        for encoder_layer_instance in self.encoder_layer:
            encoder_layer_instance.build(current_input_shape_for_layer)
        super().build(input_shape)

    def compute_output_spec(
        self, src, src_mask=None, pos_embed=None, output_attentions=False
    ):
        if not self.encoder_layer:
            if output_attentions:
                return src, None
            return src
        encoder_layer_output_spec = self.encoder_layer[0].compute_output_spec(
            hidden_states=src,
            attention_mask=src_mask,
            position_embeddings=pos_embed,
            output_attentions=output_attentions,
        )
        if output_attentions:
            return encoder_layer_output_spec
        return encoder_layer_output_spec

    def call(
        self,
        src,
        src_mask=None,
        pos_embed=None,
        output_attentions=False,
        training=None,
    ):
        current_hidden_tensor = src
        last_layer_attn_weights = None

        for encoder_layer_instance in self.encoder_layer:
            current_hidden_tensor, layer_attn_weights = encoder_layer_instance(
                hidden_states=current_hidden_tensor,
                attention_mask=src_mask,
                position_embeddings=pos_embed,
                output_attentions=output_attentions,
                training=training,
            )
            if output_attentions:
                last_layer_attn_weights = layer_attn_weights

        return current_hidden_tensor, last_layer_attn_weights

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "normalize_before": self.normalize_before,
                "encoder_hidden_dim": self.encoder_hidden_dim,
                "num_attention_heads": self.num_attention_heads,
                "dropout": self.dropout_rate,
                "layer_norm_eps": self.layer_norm_eps,
                "encoder_activation_function": self.encoder_activation_function,
                "activation_dropout": self.activation_dropout_rate,
                "encoder_ffn_dim": self.encoder_ffn_dim,
                "num_encoder_layers": self.num_encoder_layers,
                "kernel_initializer": self.kernel_initializer,
                "bias_initializer": self.bias_initializer,
            }
        )
        return config



import math

import keras
from keras import ops
from keras import random

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.causal_lm import CausalLM
from keras_hub.src.models.parseq.parseq_backbone import PARSeqBackbone
from keras_hub.src.models.parseq.parseq_causal_lm_preprocessor import (
    PARSeqCausalLMPreprocessor,
)
from keras_hub.src.utils.tensor_utils import any_equal


@keras_hub_export("keras_hub.models.PARSeqCausalLM")
class PARSeqCausalLM(CausalLM):
    """Scene Text Recognition with PARSeq.
    Performs OCR in natural scenes using the PARSeq model described in
    [Scene Text Recognition with Permuted Autoregressive Sequence Models](
    https://arxiv.org/abs/2207.06966). PARSeq is a ViT-based model that allows
    iterative decoding by performing an autoregressive decoding phase, followed
    by a refinement phase.
    Args:
        preprocessor: A `keras_hub.models.Preprocessor` instance or a
            `keras.Layer` instance. The preprocessor to use for the model.
        backbone: A `keras_hub.models.PARSeqBackbone` instance or a
            `keras.Model`. The backbone model to use for the model.
        num_perms: int. The number of permutations to generate for training.
            Defaults to 6.
        add_forward_perms: bool. Whether to add forward permutations to the
            generated permutations. Defaults to `True`.
        add_mirrored_perms: bool. Whether to add mirrored permutations to the
            generated permutations. Defaults to `True`.
        seed: int. The random seed to use for generating permutations.
            Defaults to `None`, which means no seed is set.
        **kwargs: Additional keyword arguments passed to the base
            `keras_hub.models.CausalLM` constructor.

    Examples:

    Call `predict()` to run inference.
    ```python
    # Load preset and run inference
    images = np.random.randint(0, 256, size=(2, 32, 128, 3))
    parseq = keras_hub.models.PARSeqCausalLM.from_preset(
        "parseq_vit"
    )
    parseq.generate(images)

    # Call `fit()` on a single batch.
    images = np.random.randint(0, 256, size=(2, 32, 128, 3))
    token_ids = np.array([[1, 2, 3, 4], [1, 2, 3, 0]])
    padding_mask = np.array([[1, 1, 1, 1], [1, 1, 1, 0]])
    parseq = keras_hub.models.PARSeqCausalLM.from_preset(
        "parseq_vit"
    )
    parseq.fit(
        x={
            "images": images,
            "token_ids": token_ids,
            "padding_mask": padding_mask
        },
        batch_size=2,
    )
    ```
    # Call `fit()` with custom loss, optimizer and image encoder.
    ```python
    # Initialize the image encoder, preprocessor and tokenizer
    mean, std = 0.5, 0.5
    image_converter = PARSeqImageConverter(
        image_size=(32, 128),
        offset=-mean / std,
        scale=1.0 / 255.0 / std,
        interpolation="bicubic",
    )
    tokenizer = PARSeqTokenizer(max_label_length=25)
    preprocessor = keras_hub.models.PARSeqCausalLMPreprocessor(
        image_converter=image_converter,
        tokenizer=tokenizer,
    )

    # Create the backbone
    image_encoder = ViTBackbone(
        image_shape=(32, 128, 3),
        patch_size=(4, 8),
        num_layers=12,
        num_heads=6,
        hidden_dim=384,
        mlp_dim=384 * 4,
        use_class_token=False,
        name="encoder",
    )
    backbone = PARSeqBackbone(
        vocabulary_size=97,
        max_label_length=25,
        image_encoder=image_encoder,
        num_decoder_heads=12,
        num_decoder_layers=1,
        decoder_hidden_dim=384,
        decoder_mlp_dim=4 * 384,
    )
    # Create the PARSeq model
    parseq = keras_hub.models.PARSeqCausalLM(
        backbone=backbone,
        preprocessor=preprocessor,
    )
    parseq.compile(
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=keras.optimizers.Adam(5e-5),
    )
    parseq.fit(
        x={
            "images": images,
            "token_ids": token_ids,
            "padding_mask": padding_mask
        },
        batch_size=2,
    )
    ```
    """

    backbone_cls = PARSeqBackbone
    preprocessor_cls = PARSeqCausalLMPreprocessor

    def __init__(
        self,
        preprocessor,
        backbone,
        num_perms=6,
        add_forward_perms=True,
        add_mirrored_perms=True,
        seed=None,
        end_token_id=0,  # default tokenizer.end_token_id
        **kwargs,
    ):
        # === Layers ===
        self.preprocessor = preprocessor
        self.backbone = backbone

        # === Functional Model ===
        # This must be "backbone.input" i.e. the full input structure,
        # rather than "backbone.inputs" which is the flattened list of inputs.
        inputs = backbone.input
        outputs = backbone(inputs=inputs)
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

        # === Config ===
        self.num_perms = num_perms
        self.add_forward_perms = add_forward_perms
        self.add_mirrored_perms = add_mirrored_perms
        self.end_token_id = end_token_id
        self.seed = seed
        self.seed_generator = keras.random.SeedGenerator(seed)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_perms": self.num_perms,
                "add_forward_perms": self.add_forward_perms,
                "add_mirrored_perms": self.add_mirrored_perms,
                "seed": self.seed,
                "end_token_id": self.end_token_id,
            }
        )

        return config

    def compile(
        self,
        optimizer="auto",
        loss="auto",
        *,
        weighted_metrics="auto",
        sampler="greedy",
        **kwargs,
    ):
        if loss == "auto":
            loss = keras.losses.SparseCategoricalCrossentropy(
                from_logits=True,
                ignore_class=self.preprocessor.tokenizer.pad_token_id,
            )
        super().compile(
            optimizer=optimizer,
            loss=loss,
            weighted_metrics=weighted_metrics,
            sampler=sampler,
            **kwargs,
        )

    def compute_loss(
        self, x, y, y_pred, sample_weight, training=True, *args, **kwargs
    ):
        # For keras we have fixed input for all batches, so in this case
        # we permute 23 tokens excluding BOS and EOS tokens instead of max
        # characters for current batch used in torch implementation
        # -1 because we will be generating permutation mask for considering
        # tokens before creating target label.
        max_num_chars = self.backbone.max_label_length - 1
        perms = self.generate_training_permutations(max_num_chars)
        max_label_length = self.backbone.max_label_length
        memory = self.backbone.image_encoder(x["images"])
        batch_size = ops.shape(x["images"])[0]
        losses = []
        for i in range(ops.shape(perms)[0]):
            query_mask, content_mask = self.generate_attention_masks(perms[i])
            query_mask = ops.broadcast_to(
                query_mask, (batch_size, max_label_length, max_label_length)
            )
            content_mask = ops.broadcast_to(
                content_mask, (batch_size, max_label_length, max_label_length)
            )
            out = self.backbone.decoder(
                x["token_ids"],
                memory,
                padding_mask=x["padding_mask"],
                query_mask=query_mask,
                content_mask=content_mask,
            )
            y_pred = self.backbone.head(out)
            loss = super().compute_loss(
                x=x, y=y, y_pred=y_pred, sample_weight=sample_weight, **kwargs
            )
            losses.append(loss)
            if i == 1:
                # Sample weights are set to zero for end-of-sequence (EOS)
                # tokens to prevent them from affecting loss calculations.
                # reference: https://github.com/baudm/parseq/blob/1902db043c029a7e03a3818c616c06600af574be/strhub/models/parseq/system.py#L194 # noqa: E501
                sample_weight = ops.logical_and(
                    y != self.end_token_id, sample_weight
                )

        return ops.sum(losses) / ops.shape(perms)[0]

    def generate_training_permutations(self, max_num_chars):
        max_gen_perms = (
            self.num_perms // 2 if self.add_mirrored_perms else self.num_perms
        )

        if max_num_chars == 1:
            return ops.expand_dims(ops.arange(3), axis=0)

        perms = [ops.arange(max_num_chars)] if self.add_forward_perms else []
        max_num_perms = math.factorial(max_num_chars)
        max_gen_perms = min(max_gen_perms, max_num_perms)

        for _ in range(max_gen_perms - len(perms)):
            perm = random.shuffle(
                ops.arange(max_num_chars), seed=self.seed_generator
            )
            perms.append(perm)

        perms = ops.stack(perms)
        comp = ops.flip(perms, axis=-1)
        perms = ops.stack([perms, comp])
        perms = ops.reshape(
            ops.transpose(perms, (1, 0, 2)), (-1, max_num_chars)
        )

        bos_idx = ops.zeros((ops.shape(perms)[0], 1), dtype="int32")
        eos_idx = ops.full(
            (ops.shape(perms)[0], 1), max_num_chars + 1, dtype="int32"
        )
        perms = ops.concatenate([bos_idx, perms + 1, eos_idx], axis=1)

        if perms.shape[0] > 1:
            perms = ops.scatter_update(
                perms,
                ops.concatenate(
                    [
                        ops.ones((max_num_chars + 1, 1), dtype="int32"),
                        ops.expand_dims(
                            ops.arange(1, max_num_chars + 2, dtype="int32"),
                            axis=1,
                        ),
                    ],
                    axis=1,
                ),
                max_num_chars + 1 - ops.arange(max_num_chars + 1),
            )

        return perms

    def generate_attention_masks(self, perm):
        """Generate attention masks given a sequence permutation
        (includes pos. for BOS and EOS tokens)"""
        input_length = ops.shape(perm)[0]
        mask = ops.ones((input_length, input_length))
        for i in range(input_length - 1):
            masked_keys = perm[i + 1 : input_length]
            query_idx = ops.broadcast_to(perm[i], ops.shape(masked_keys))
            indices = ops.stack((query_idx, masked_keys), axis=1)
            mask = keras.ops.scatter_update(
                mask, indices, keras.ops.zeros(ops.shape(masked_keys)[0])
            )
        content_mask = mask[:-1, :-1]
        mask = mask * (1 - ops.eye(input_length))
        query_mask = mask[1:, :-1]
        return query_mask, content_mask

    def call_with_cache(
        self,
        token_ids,
        cache,
        cache_update_index,
        img_embeddings,
        padding_mask=None,
    ):
        bs = ops.shape(token_ids)[0]
        # <bos> stands for the null context. We only supply position information
        # for characters after <bos>.
        content = ops.where(
            cache_update_index == 0,
            self.backbone.decoder_hidden_dim**0.5
            * self.backbone.decoder.token_embedding(token_ids),
            ops.expand_dims(
                self.backbone.decoder.pos_query_embeddings[
                    :, cache_update_index - 1, :
                ],
                axis=0,
            )
            + self.backbone.decoder_hidden_dim**0.5
            * self.backbone.decoder.token_embedding(token_ids),
        )
        content = self.backbone.decoder.dropout(content)

        query = ops.ones((bs, 1, 1)) * ops.expand_dims(
            self.backbone.decoder.pos_query_embeddings[
                :, cache_update_index, :
            ],
            axis=0,
        )
        query = self.backbone.decoder.dropout(query)

        query_cache = []
        content_cache = []
        for i, decoder_layer in enumerate(self.backbone.decoder.decoder_layers):
            last = i == self.backbone.num_decoder_layers - 1
            current_query_cache = cache[:, i, 0, ...]
            current_content_cache = cache[:, i, 1, ...]
            (
                query,
                content,
                query_self_attention_new_cache,
                content_self_attention_cache,
            ) = decoder_layer(
                query=query,
                content=content,
                memory=img_embeddings,
                padding_mask=padding_mask,
                update_content=not last,
                query_self_attention_cache=current_query_cache,
                query_self_attention_cache_update_index=cache_update_index,
                content_self_attention_cache=current_content_cache,
                content_self_attention_cache_update_index=cache_update_index,
            )
            query_cache.append(query_self_attention_new_cache)
            content_cache.append(content_self_attention_cache)

        query_cache = ops.stack(query_cache, axis=1)
        content_cache = ops.stack(content_cache, axis=1)
        cache = ops.stack([query_cache, content_cache], axis=2)
        hidden_states = self.backbone.decoder.layer_norm(query)
        logits = self.backbone.head(hidden_states)
        return logits, hidden_states, cache

    def _build_cache(self, token_ids, img_embeddings, padding_mask):
        batch_size = ops.shape(token_ids)[0]
        max_length = ops.shape(token_ids)[1]
        num_layers = self.backbone.num_decoder_layers
        head_dim = (
            self.backbone.decoder_hidden_dim // self.backbone.num_decoder_heads
        )
        num_heads = self.backbone.num_decoder_heads
        shape = [batch_size, num_layers, 2, 2, max_length, num_heads, head_dim]
        cache = ops.zeros(shape)

        # Seed the cache.
        logits, hidden_states, cache = self.call_with_cache(
            token_ids=token_ids,
            img_embeddings=img_embeddings,
            cache=cache,
            cache_update_index=0,
            padding_mask=padding_mask,
        )
        return hidden_states, cache

    def generate_step(self, inputs, stop_token_ids=None):
        token_ids, padding_mask, images = (
            inputs["token_ids"],
            inputs["padding_mask"],
            inputs["images"],
        )
        images_shape = ops.shape(images)
        if len(images_shape) == 3:
            # Handle an unbatched image. Unlike `token_ids` and `padding_mask`
            # this will not automatically be upranked.
            images = ops.expand_dims(images, axis=0)

        img_embeddings = self.backbone.image_encoder(images)
        # Create and seed cache with a single forward pass.
        hidden_states, cache = self._build_cache(
            token_ids=token_ids,
            img_embeddings=img_embeddings,
            padding_mask=padding_mask,
        )
        # Compute the lengths of all user inputted tokens ids.
        row_lengths = ops.sum(ops.cast(padding_mask, "int32"), axis=-1)
        # Start at the first index that has no user inputted id.
        index = ops.min(row_lengths)

        def next(prompt, cache, index):
            # The cache index is the index of our previous token.
            cache_update_index = index - 1
            batch_size = ops.shape(prompt)[0]
            prompt = ops.slice(prompt, [0, index - 1], [batch_size, 1])
            logits, hidden_states, cache = self.call_with_cache(
                token_ids=prompt,
                cache=cache,
                cache_update_index=cache_update_index,
                img_embeddings=img_embeddings,
            )
            return (
                ops.squeeze(logits, axis=1),
                ops.squeeze(hidden_states, axis=1),
                cache,
            )

        token_ids = self.sampler(
            next=next,
            prompt=token_ids,
            cache=cache,
            index=index,
            mask=padding_mask,
            stop_token_ids=stop_token_ids,
            hidden_states=hidden_states,
            model=self,
        )

        # Compute an output padding mask with the token ids we updated.
        if stop_token_ids is not None:
            # Build a mask of `stop_token_ids` locations not in the original
            # prompt (not in locations where `padding_mask` is True).
            end_locations = any_equal(
                token_ids, stop_token_ids, ops.logical_not(padding_mask)
            )

            end_locations = ops.cast(end_locations, "int32")
            # Use cumsum to get ones in all locations after end_locations.
            cumsum = ops.cast(ops.cumsum(end_locations, axis=-1), "int32")
            overflow = cumsum - end_locations
            # Our padding mask is the inverse of these overflow locations.
            padding_mask = ops.logical_not(ops.cast(overflow, "bool"))
        else:
            # Without early stopping, all locations will have been updated.
            padding_mask = ops.ones_like(token_ids, dtype="bool")
        return {
            "token_ids": token_ids,
            "padding_mask": padding_mask,
            "images": images,
        }



import keras

from keras_hub.src.layers.modeling.rotary_embedding import RotaryEmbedding
from keras_hub.src.utils.keras_utils import clone_initializer


def moonshine_kernel_initializer(initializer_range=0.02):
    return keras.initializers.TruncatedNormal(stddev=initializer_range)


@keras.saving.register_keras_serializable(package="keras_hub")
class MoonshineRotaryEmbedding(RotaryEmbedding):
    """
    Moonshine rotary embedding layer.

    Computes rotary positional embeddings using precomputed inverse frequencies
    for a fraction of dimensions.

    The layer stores inverse frequency weights as a non-trainable parameter and
    computes sinusoidal embeddings based on input positions. Unlike KerasHub's
    `RotaryEmbedding` class, this implementation explicitly requires `head_dim`
    and applies `partial_rotary_factor` for selective rotary embedding, whereas
    KerasHub uses `max_wavelength` without partial application.

    Args:
        head_dim: int. The dimensionality of each attention head, determining
            the feature space for rotary embeddings.
        max_position_embeddings: int, optional. The maximum sequence length the
            model can process, controlling the positional embedding scale.
            Defaults to 2048.
        base_value: float, optional. Base value for computing inverse
            frequencies. Higher values result in longer wavelengths. Defaults to
            10000.
        partial_rotary_factor: float, optional. The fraction of `head_dim`
            dimensions that receive rotary embeddings, balancing rotary and
            non-rotary components. Defaults to 0.62.
        dtype: string, optional. The data type for model computations and
            weights. Defaults to None.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    # References:
    # Based on the UsefulSensors implementation of the RotaryEmbedding class (https://github.com/usefulsensors/moonshine/blob/4a000427bd36a1c2c6d20a86c672dbd850b44c88/moonshine/model.py#L176-L193).

    def __init__(
        self,
        head_dim,
        max_position_embeddings=2048,
        base_value=10000,
        partial_rotary_factor=0.62,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.head_dim = head_dim
        self.max_position_embeddings = max_position_embeddings
        self.base_value = base_value
        self.partial_rotary_factor = partial_rotary_factor
        self.built = False
        self.rotary_dim = None
        self.inv_freq = None

    def build(self, input_shape):
        if self.built:
            return
        # Create and track the non-trainable weight immediately.
        rotary_dim = int(self.head_dim * self.partial_rotary_factor)
        rotary_dim = (rotary_dim // 2) * 2
        if rotary_dim <= 0:
            raise ValueError(
                f"Calculated rotary_dim ({rotary_dim}) must be a positive even "
                f"number. Check head_dim ({self.head_dim}) and "
                f"partial_rotary_factor ({self.partial_rotary_factor})."
            )
        self.rotary_dim = rotary_dim
        rotary_dim_half = rotary_dim // 2

        # Compute inv_freq.
        inv_freq = 1.0 / (
            self.base_value
            ** (
                keras.ops.arange(0, rotary_dim_half, dtype=self.dtype)
                / rotary_dim_half
            )
        )

        # Set the non-trainable weight using the computed tensor.
        self.inv_freq = self.add_weight(
            name="inv_freq",
            shape=(rotary_dim_half,),
            initializer=keras.initializers.Constant(inv_freq),
            trainable=False,
            dtype=self.dtype,
        )
        self.built = True

    def call(self, t):
        t_cast = keras.ops.cast(t, keras.ops.dtype(self.inv_freq))
        freqs = keras.ops.einsum("i,j->ij", t_cast, self.inv_freq)
        emb = keras.ops.stack((freqs, freqs), axis=-1)
        shape_list = list(keras.ops.shape(emb))
        shape_list[-2:] = [-1]
        return keras.ops.reshape(emb, shape_list)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "head_dim": self.head_dim,
                "max_position_embeddings": self.max_position_embeddings,
                "base_value": self.base_value,
                "partial_rotary_factor": self.partial_rotary_factor,
                "dtype": self.dtype,
            }
        )
        return config


@keras.saving.register_keras_serializable(package="keras_hub")
class MoonshineMLP(keras.layers.Layer):
    """
    Moonshine MLP layer.

    Implements a Multi-Layer Perceptron (MLP) for Moonshine models with support
    for both `SwiGLU` and `LinearGeLU` activation patterns. The MLP consists of
    two dense layers with an activation function in between, expanding the input
    dimension before projecting back to the original dimension.

    Args:
        hidden_dim: int. The dimensionality of the input and output tensors.
        feedforward_expansion_factor: float. The factor by which to expand the
            hidden dimension in the intermediate layer.
        use_swiglu_activation: bool, optional. If `True`, uses SwiGLU activation
            (SiLU with gating). If `False`, uses standard GeLU activation.
            Defaults to `True`.
        initializer_range: float, optional. The standard deviation for kernel
            initialization. Defaults to 0.02.
        dtype: string, optional. The data type for model computations and
            weights. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    # References:
    # Based on the HuggingFace implementation of the MoonshineEncoderMLP and
    # MoonshineDecoderMLP classes (https://github.com/huggingface/transformers/blob/fc8764c9a618add64c33e83720f974750bcd0978/src/transformers/models/moonshine/modeling_moonshine.py#L66-L94).

    def __init__(
        self,
        hidden_dim,
        feedforward_expansion_factor,
        use_swiglu_activation=True,
        initializer_range=0.02,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.hidden_dim = hidden_dim
        self.feedforward_expansion_factor = feedforward_expansion_factor
        self.use_swiglu_activation = use_swiglu_activation
        self.kernel_initializer = moonshine_kernel_initializer(
            initializer_range=initializer_range
        )
        self.initializer_range = initializer_range

        if use_swiglu_activation:
            # First dense layer produces (2 * feedforward_expansion_factor *
            # hidden_dim) outputs.
            self.dense_1 = keras.layers.Dense(
                int(hidden_dim * feedforward_expansion_factor * 2),
                use_bias=True,
                name="dense_1",
                dtype=self.dtype,
                kernel_initializer=clone_initializer(self.kernel_initializer),
            )
            # Activation layer using "silu" (Swish activation).
            self.activation = keras.layers.Activation(
                "silu", name="activation", dtype=self.dtype
            )
        else:
            # Taken from pretrained weights.
            # First dense layer: output dimension is (hidden_dim *
            # feedforward_expansion_factor).
            self.dense_1 = keras.layers.Dense(
                int(hidden_dim * feedforward_expansion_factor),
                use_bias=True,
                name="dense_1",
                dtype=self.dtype,
                kernel_initializer=clone_initializer(self.kernel_initializer),
            )
            self.activation = keras.layers.Activation(
                "gelu", name="activation", dtype=self.dtype
            )

        # Second dense layer projects back to hidden_dim.
        self.dense_2 = keras.layers.Dense(
            hidden_dim,
            use_bias=True,
            name="dense_2",
            dtype=self.dtype,
            kernel_initializer=clone_initializer(self.kernel_initializer),
        )

    def build(self, input_shape):
        super().build(input_shape)
        # Build the first dense layer using the original input shape.
        self.dense_1.build(input_shape)
        # After dense_1, the output shape becomes: (..., 2 *
        # feedforward_expansion_factor * hidden_dim).
        # When splitting, each part will have shape (...,
        # feedforward_expansion_factor * hidden_dim).
        new_input_shape = list(input_shape)
        new_input_shape[-1] = (
            self.hidden_dim * self.feedforward_expansion_factor
        )
        self.dense_2.build(tuple(new_input_shape))

    def call(self, inputs):
        x = self.dense_1(inputs)
        if self.use_swiglu_activation:
            x1, gate = keras.ops.split(x, 2, axis=-1)
            activated_gate = self.activation(gate)
            x = x1 * activated_gate
        else:
            x = self.activation(x)
        output = self.dense_2(x)
        return output

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "feedforward_expansion_factor": self.feedforward_expansion_factor,  # noqa: E501
                "use_swiglu_activation": self.use_swiglu_activation,
                "initializer_range": self.initializer_range,
                "dtype": self.dtype,
            }
        )
        return config



"""Mistral model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "mistral_7b_en": {
        "metadata": {
            "description": "Mistral 7B base model",
            "params": 7241732096,
            "path": "mistral",
        },
        "kaggle_handle": "kaggle://keras/mistral/keras/mistral_7b_en/8",
    },
    "mistral_0.3_7b_en": {
        "metadata": {
            "description": "Mistral 7B base version 0.3 model",
            "params": 7248023552,
            "path": "mistral",
        },
        "kaggle_handle": "kaggle://keras/mistral/keras/mistral_0.3_7b_en/1",
    },
    "mistral_instruct_7b_en": {
        "metadata": {
            "description": "Mistral 7B instruct model",
            "params": 7241732096,
            "path": "mistral",
        },
        "kaggle_handle": "kaggle://keras/mistral/keras/mistral_instruct_7b_en/8",
    },
    "mistral_0.2_instruct_7b_en": {
        "metadata": {
            "description": "Mistral 7B instruct version 0.2 model",
            "params": 7241732096,
            "path": "mistral",
        },
        "kaggle_handle": "kaggle://keras/mistral/keras/mistral_0.2_instruct_7b_en/3",
    },
    "mistral_0.3_instruct_7b_en": {
        "metadata": {
            "description": "Mistral 7B instruct version 0.3 model",
            "params": 7248023552,
            "path": "mistral",
        },
        "kaggle_handle": "kaggle://keras/mistral/keras/mistral_0.3_instruct_7b_en/1",
    },
}



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.models.mobilenetv5.mobilenetv5_backbone import (
    MobileNetV5Backbone,
)


@keras_hub_export("keras_hub.layers.MobileNetV5ImageConverter")
class MobileNetV5ImageConverter(ImageConverter):
    backbone_cls = MobileNetV5Backbone



from keras_hub.src.models.qwen3.qwen3_backbone import Qwen3Backbone
from keras_hub.src.models.qwen3.qwen3_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, Qwen3Backbone)



from keras_hub.src.models.seq_2_seq_lm import Seq2SeqLM


class AudioToText(Seq2SeqLM):
    """Base class for audio-to-text models.

    `AudioToText` tasks wrap a `keras_hub.models.Backbone` (capable of
    processing audio and text features) and a
    `keras_hub.models.AudioToTextPreprocessor` to create a model for
    audio-to-text tasks like speech recognition or audio transcription.

    These models typically consist of an encoder that processes audio input
    and a decoder that generates a textual representation.

    `AudioToText` tasks provide a high-level `generate()` method for
    auto-regressively generating text from audio input. An optional text
    prompt can also be provided to the decoder to guide generation. The
    sampling strategy for generation (e.g., greedy, top-k, top-p) can be
    controlled via the `sampler` argument in the `compile()` method.

    When calling `fit()`, inputs should consist of audio data and corresponding
    target text transcriptions. The model is trained to predict the target text
    token-by-token.

    All `AudioToText` tasks include a `from_preset()` constructor which
    can be used to load pre-trained configurations and weights for specific
    audio-to-text models.
    This constructor can also be called on the base `AudioToText` class,
    which will automatically select the correct subclass based on the preset.

    Examples:
    ```python
    # Load a Moonshine backbone with pre-trained weights.
    # AudioToText is a base class. You will typically work with a specific
    # implementation, such as `keras_hub.models.MoonshineAudioToText`.
    # The following examples demonstrate common usage patterns.

    # Initialize a model from a preset using the specific subclass.
    audio_to_text = keras_hub.models.MoonshineAudioToText.from_preset(
        "moonshine_base_en"
    )

    # Initialize a model from a preset using the base class.
    audio_to_text_model_base = keras_hub.models.AudioToText.from_preset(
        "moonshine_base_en"
    )

    # Generate text from an audio input.
    audio_input_tensor = keras.random.normal((1, 16000, 1))
    generated_output = audio_to_text_model.generate(
        {"audio": audio_input_tensor}
    )

    # Generate conditioned on the `"The quick brown fox."` as an input sequence.
    prompted_output = audio_to_text_model.generate(
        {"audio": audio_input_tensor, "text": "The quick brown fox."}
    )

    # Use a different sampling strategy for generation.
    audio_to_text_model.compile(sampler="greedy")
    greedy_output = audio_to_text_model.generate(
        {"audio": audio_input_tensor}
    )
    """

    # TODO: Fill in once audio to text task model requirements are clearer.



import keras
import numpy as np

from keras_hub.src.models.d_fine.d_fine_utils import inverse_sigmoid


class DFineGate(keras.layers.Layer):
    """Gating layer for combining two input tensors using learnable gates.

    This layer is used within the `DFineDecoderLayer` to merge the output of
    the self-attention mechanism (residual) with the output of the
    cross-attention mechanism (`hidden_states`). It computes a weighted sum of
    the two inputs, where the weights are learned gates. The result is
    normalized using layer normalization.

    Args:
        hidden_dim: int, The hidden dimension size for the gate computation.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(self, hidden_dim, dtype=None, **kwargs):
        super().__init__(dtype=dtype, **kwargs)
        self.hidden_dim = hidden_dim
        self.norm = keras.layers.LayerNormalization(
            epsilon=1e-5, name="norm", dtype=self.dtype_policy
        )
        self.gate = keras.layers.Dense(
            2 * self.hidden_dim,
            name="gate",
            dtype=self.dtype_policy,
            kernel_initializer="zeros",
            bias_initializer="zeros",
        )

    def build(self, input_shape):
        batch_dim, seq_len_dim = None, None
        if input_shape and len(input_shape) == 3:
            batch_dim = input_shape[0]
            seq_len_dim = input_shape[1]
        gate_build_shape = (batch_dim, seq_len_dim, 2 * self.hidden_dim)
        self.gate.build(gate_build_shape)
        norm_build_shape = (batch_dim, seq_len_dim, self.hidden_dim)
        self.norm.build(norm_build_shape)
        super().build(input_shape)

    def call(self, second_residual, hidden_states, training=None):
        gate_input = keras.ops.concatenate(
            [second_residual, hidden_states], axis=-1
        )
        gates_linear_output = self.gate(gate_input)
        gates = keras.ops.sigmoid(gates_linear_output)
        gate_chunks = keras.ops.split(gates, 2, axis=-1)
        gate1 = gate_chunks[0]
        gate2 = gate_chunks[1]
        gated_sum = gate1 * second_residual + gate2 * hidden_states
        hidden_states = self.norm(gated_sum, training=training)
        return hidden_states

    def get_config(self):
        config = super().get_config()
        config.update({"hidden_dim": self.hidden_dim})
        return config


class DFineMLP(keras.layers.Layer):
    """Multi-layer perceptron (MLP) layer.

    This layer implements a standard MLP. It is used in several places within
    the D-FINE model, such as the `reg_conf` head inside `DFineLQE` for
    predicting quality scores and the `pre_bbox_head` in `DFineDecoder` for
    initial bounding box predictions.

    Args:
        input_dim: int, The input dimension.
        hidden_dim: int, The hidden dimension for intermediate layers.
        output_dim: int, The output dimension.
        num_layers: int, The number of layers in the MLP.
        activation_function: str, The activation function to use between layers.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
        last_layer_initializer: str or Initializer, optional, Special
            initializer for the final layer's weights and biases. If `None`,
            uses `kernel_initializer` and `bias_initializer`. Defaults to
            `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        output_dim,
        num_layers,
        activation_function="relu",
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        last_layer_initializer=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.num_layers = num_layers
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.activation_function = activation_function
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        # NOTE: In the original code, this is done by searching the modules for
        # specific last layers, instead, we find the last layer in each of the
        # specific modules with `num_layers - 1`.
        self.last_layer_initializer = keras.initializers.get(
            last_layer_initializer
        )
        h = [hidden_dim] * (num_layers - 1)
        input_dims = [input_dim] + h
        output_dims = h + [output_dim]
        self.dense_layers = []
        for i, (_, out_dim) in enumerate(zip(input_dims, output_dims)):
            # NOTE: Req. for handling the case of initializing the final layers'
            # weights and biases to zero when required (for ex: `bbox_embed` or
            # `reg_conf`).
            is_last_layer = i == num_layers - 1
            current_kernel_init = self.kernel_initializer
            current_bias_init = self.bias_initializer
            if is_last_layer and self.last_layer_initializer is not None:
                current_kernel_init = self.last_layer_initializer
                current_bias_init = self.last_layer_initializer
            self.dense_layers.append(
                keras.layers.Dense(
                    units=out_dim,
                    name=f"mlp_dense_layer_{i}",
                    dtype=self.dtype_policy,
                    kernel_initializer=current_kernel_init,
                    bias_initializer=current_bias_init,
                )
            )
        self.activation_layer = keras.layers.Activation(
            activation_function,
            name="mlp_activation_layer",
            dtype=self.dtype_policy,
        )

    def build(self, input_shape):
        if self.dense_layers:
            current_build_shape = input_shape
            for i, dense_layer in enumerate(self.dense_layers):
                dense_layer.build(current_build_shape)
                current_build_shape = dense_layer.compute_output_shape(
                    current_build_shape
                )
        super().build(input_shape)

    def call(self, stat_features, training=None):
        x = stat_features
        for i in range(self.num_layers):
            dense_layer = self.dense_layers[i]
            x = dense_layer(x)
            if i < self.num_layers - 1:
                x = self.activation_layer(x)
        return x

    def compute_output_spec(self, stat_features_spec):
        output_shape = list(stat_features_spec.shape)
        output_shape[-1] = self.output_dim
        return keras.KerasTensor(
            shape=tuple(output_shape), dtype=self.compute_dtype
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_dim": self.input_dim,
                "hidden_dim": self.hidden_dim,
                "output_dim": self.output_dim,
                "num_layers": self.num_layers,
                "activation_function": self.activation_function,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "last_layer_initializer": keras.initializers.serialize(
                    self.last_layer_initializer
                ),
            }
        )
        return config


class DFineSourceFlattener(keras.layers.Layer):
    """Layer to flatten and concatenate a list of source tensors.

    This layer is used in `DFineBackbone` to process feature maps from the
    `DFineHybridEncoder`. It takes a list of multi-scale feature maps,
    flattens each along its spatial dimensions, and concatenates them
    along the sequence dimension.

    Args:
        channel_axis: int, optional, The channel axis. Defaults to `None`.
        data_format: str, optional, The data format. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self, channel_axis=None, data_format=None, dtype=None, **kwargs
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.channel_axis = channel_axis
        self.data_format = data_format

    def call(self, sources, training=None):
        source_flatten = []
        for i, source_item in enumerate(sources):
            if self.data_format == "channels_first":
                source_item = keras.ops.transpose(source_item, [0, 2, 3, 1])
            batch_size = keras.ops.shape(source_item)[0]
            channels = keras.ops.shape(source_item)[-1]
            source_reshaped = keras.ops.reshape(
                source_item, (batch_size, -1, channels)
            )
            source_flatten.append(source_reshaped)
        source_flatten_concatenated = keras.ops.concatenate(
            source_flatten, axis=1
        )
        return source_flatten_concatenated

    def compute_output_shape(self, sources_shape):
        if not sources_shape or not isinstance(sources_shape, list):
            return tuple()
        if not all(isinstance(s, tuple) and len(s) == 4 for s in sources_shape):
            return tuple()
        batch_size = sources_shape[0][0]
        if self.data_format == "channels_first":
            channels = sources_shape[0][1]
        else:
            channels = sources_shape[0][-1]
        calculated_spatial_elements = []
        for s_shape in sources_shape:
            if self.data_format == "channels_first":
                h, w = s_shape[2], s_shape[3]
            else:
                h, w = s_shape[1], s_shape[2]
            if h is None or w is None:
                calculated_spatial_elements.append(None)
            else:
                calculated_spatial_elements.append(h * w)
        if any(elem is None for elem in calculated_spatial_elements):
            total_spatial_elements = None
        else:
            total_spatial_elements = sum(calculated_spatial_elements)
        return (batch_size, total_spatial_elements, channels)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "channel_axis": self.channel_axis,
                "data_format": self.data_format,
            }
        )
        return config


class DFineContrastiveDenoisingGroupGenerator(keras.layers.Layer):
    """Layer to generate denoising groups for contrastive learning.

    This layer, used in `DFineBackbone`, implements the core logic for
    contrastive denoising, a key training strategy in D-FINE. It takes ground
    truth `targets`, adds controlled noise to labels and boxes, and generates
    the necessary attention masks, queries, and reference points for the
    decoder. Due to functional model constraints, noise is generated once at
    model initialization.

    Args:
        num_labels: int, The number of object classes.
        num_denoising: int, The number of denoising queries.
        label_noise_ratio: float, The ratio of label noise to apply.
        box_noise_scale: float, The scale of box noise to apply.
        seed: int, optional, The random seed for noise generation.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        num_labels,
        num_denoising,
        label_noise_ratio,
        box_noise_scale,
        seed=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.num_labels = num_labels
        self.num_denoising = num_denoising
        self.label_noise_ratio = label_noise_ratio
        self.box_noise_scale = box_noise_scale
        self.seed = seed
        self.seed_generator = keras.random.SeedGenerator(seed)

    def build(self, input_shape):
        super().build(input_shape)

    def call(self, targets, num_queries):
        if self.num_denoising <= 0:
            return None, None, None, None
        num_ground_truths = [len(t["labels"]) for t in targets]
        max_gt_num = 0
        if num_ground_truths:
            max_gt_num = max(num_ground_truths)
        if max_gt_num == 0:
            return None, None, None, None
        num_groups_denoising_queries = self.num_denoising // max_gt_num
        num_groups_denoising_queries = (
            1
            if num_groups_denoising_queries == 0
            else num_groups_denoising_queries
        )
        batch_size = len(num_ground_truths)
        input_query_class = []
        input_query_bbox = []
        pad_gt_mask = []
        for i in range(batch_size):
            num_gt = num_ground_truths[i]
            if num_gt > 0:
                labels = targets[i]["labels"]
                boxes = targets[i]["boxes"]
                padded_class_labels = keras.ops.pad(
                    labels,
                    [[0, max_gt_num - num_gt]],
                    constant_values=self.num_labels,
                )
                padded_boxes = keras.ops.pad(
                    keras.ops.cast(boxes, dtype=self.compute_dtype),
                    [[0, max_gt_num - num_gt], [0, 0]],
                    constant_values=0.0,
                )
                mask = keras.ops.concatenate(
                    [
                        keras.ops.ones([num_gt], dtype="bool"),
                        keras.ops.zeros([max_gt_num - num_gt], dtype="bool"),
                    ]
                )
            else:
                padded_class_labels = keras.ops.full(
                    [max_gt_num], self.num_labels, dtype="int32"
                )
                padded_boxes = keras.ops.zeros(
                    [max_gt_num, 4], dtype=self.compute_dtype
                )
                mask = keras.ops.zeros([max_gt_num], dtype="bool")
            input_query_class.append(padded_class_labels)
            input_query_bbox.append(padded_boxes)
            pad_gt_mask.append(mask)
        input_query_class = keras.ops.stack(input_query_class, axis=0)
        input_query_bbox = keras.ops.stack(input_query_bbox, axis=0)
        pad_gt_mask = keras.ops.stack(pad_gt_mask, axis=0)
        input_query_class = keras.ops.tile(
            input_query_class, [1, 2 * num_groups_denoising_queries]
        )
        input_query_bbox = keras.ops.tile(
            input_query_bbox, [1, 2 * num_groups_denoising_queries, 1]
        )
        pad_gt_mask = keras.ops.tile(
            pad_gt_mask, [1, 2 * num_groups_denoising_queries]
        )
        negative_gt_mask = keras.ops.zeros(
            [batch_size, max_gt_num * 2, 1], dtype=self.compute_dtype
        )
        updates_neg = keras.ops.ones(
            [batch_size, max_gt_num, 1], dtype=negative_gt_mask.dtype
        )
        negative_gt_mask = keras.ops.slice_update(
            negative_gt_mask, [0, max_gt_num, 0], updates_neg
        )
        negative_gt_mask = keras.ops.tile(
            negative_gt_mask, [1, num_groups_denoising_queries, 1]
        )
        positive_gt_mask_float = 1.0 - negative_gt_mask
        squeezed_positive_gt_mask = keras.ops.squeeze(
            positive_gt_mask_float, axis=-1
        )
        positive_gt_mask = squeezed_positive_gt_mask * keras.ops.cast(
            pad_gt_mask, dtype=squeezed_positive_gt_mask.dtype
        )
        denoise_positive_idx = []
        for i in range(batch_size):
            mask_i = positive_gt_mask[i]
            idx = keras.ops.nonzero(mask_i)[0]
            denoise_positive_idx.append(idx)
        if self.label_noise_ratio > 0:
            noise_mask = keras.random.uniform(
                keras.ops.shape(input_query_class),
                dtype=self.compute_dtype,
                seed=self.seed_generator,
            ) < (self.label_noise_ratio * 0.5)
        max_len = 0
        for idx in denoise_positive_idx:
            current_len = keras.ops.shape(idx)[0]
            if current_len > max_len:
                max_len = current_len
        padded_indices = []
        for idx in denoise_positive_idx:
            current_len = keras.ops.shape(idx)[0]
            pad_len = max_len - current_len
            padded = keras.ops.pad(idx, [[0, pad_len]], constant_values=-1)
            padded_indices.append(padded)
        dn_positive_idx = (
            keras.ops.stack(padded_indices, axis=0) if padded_indices else None
        )
        if self.label_noise_ratio > 0:
            noise_mask = keras.ops.cast(noise_mask, "bool")
            new_label = keras.random.randint(
                keras.ops.shape(input_query_class),
                0,
                self.num_labels,
                seed=self.seed_generator,
                dtype="int32",
            )
            input_query_class = keras.ops.where(
                noise_mask & pad_gt_mask,
                new_label,
                input_query_class,
            )
        if self.box_noise_scale > 0:
            known_bbox = keras.utils.bounding_boxes.convert_format(
                input_query_bbox,
                source="center_xywh",
                target="xyxy",
                dtype=self.compute_dtype,
            )
            width_height = input_query_bbox[..., 2:]
            diff = (
                keras.ops.tile(width_height, [1, 1, 2])
                * 0.5
                * self.box_noise_scale
            )
            rand_int_sign = keras.random.randint(
                keras.ops.shape(input_query_bbox),
                0,
                2,
                seed=self.seed_generator,
            )
            rand_sign = (
                keras.ops.cast(rand_int_sign, dtype=diff.dtype) * 2.0 - 1.0
            )
            rand_part = keras.random.uniform(
                keras.ops.shape(input_query_bbox),
                seed=self.seed_generator,
                dtype=self.compute_dtype,
            )
            rand_part = (rand_part + 1.0) * negative_gt_mask + rand_part * (
                1 - negative_gt_mask
            )
            rand_part = rand_part * rand_sign
            known_bbox = known_bbox + rand_part * diff
            known_bbox = keras.ops.clip(known_bbox, 0.0, 1.0)
            input_query_bbox = keras.utils.bounding_boxes.convert_format(
                known_bbox,
                source="xyxy",
                target="center_xywh",
                dtype=self.compute_dtype,
            )
        input_query_bbox = inverse_sigmoid(input_query_bbox)
        num_denoising_total = max_gt_num * 2 * num_groups_denoising_queries
        target_size = num_denoising_total + num_queries
        attn_mask = keras.ops.zeros(
            [target_size, target_size], dtype=self.compute_dtype
        )
        updates_attn1 = keras.ops.ones(
            [
                target_size - num_denoising_total,
                num_denoising_total,
            ],
            dtype=attn_mask.dtype,
        )
        attn_mask = keras.ops.slice_update(
            attn_mask, [num_denoising_total, 0], updates_attn1
        )
        for i in range(num_groups_denoising_queries):
            start = max_gt_num * 2 * i
            end = max_gt_num * 2 * (i + 1)
            updates_attn2 = keras.ops.ones(
                [end - start, start], dtype=attn_mask.dtype
            )
            attn_mask = keras.ops.slice_update(
                attn_mask, [start, 0], updates_attn2
            )
            updates_attn3 = keras.ops.ones(
                [end - start, num_denoising_total - end],
                dtype=attn_mask.dtype,
            )
            attn_mask = keras.ops.slice_update(
                attn_mask, [start, end], updates_attn3
            )
        if dn_positive_idx is not None:
            denoising_meta_values = {
                "dn_positive_idx": dn_positive_idx,
                "dn_num_group": keras.ops.convert_to_tensor(
                    num_groups_denoising_queries, dtype="int32"
                ),
                "dn_num_split": keras.ops.convert_to_tensor(
                    [num_denoising_total, num_queries], dtype="int32"
                ),
            }
        return (
            input_query_class,
            input_query_bbox,
            attn_mask,
            denoising_meta_values,
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_labels": self.num_labels,
                "num_denoising": self.num_denoising,
                "label_noise_ratio": self.label_noise_ratio,
                "box_noise_scale": self.box_noise_scale,
                "seed": self.seed,
            }
        )
        return config


class DFineAnchorGenerator(keras.layers.Layer):
    """Layer to generate anchor boxes for object detection.

    This layer is used in `DFineBackbone` to generate anchor proposals. These
    anchors are combined with the output of the encoder's bounding box head
    (`enc_bbox_head`) to create initial reference points for the decoder's
    queries.

    Args:
        anchor_image_size: tuple, The size of the input image `(height, width)`.
        feat_strides: list, The strides of the feature maps.
        data_format: str, The data format of the image channels. Can be either
            `"channels_first"` or `"channels_last"`. If `None` is specified,
            it will use the `image_data_format` value found in your Keras
            config file at `~/.keras/keras.json`. Defaults to `None`.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        anchor_image_size,
        feat_strides,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.anchor_image_size = anchor_image_size
        self.feat_strides = feat_strides
        self.data_format = data_format

    def call(self, sources_for_shape_derivation=None, grid_size=0.05):
        spatial_shapes = None
        if sources_for_shape_derivation is not None:
            if self.data_format == "channels_first":
                spatial_shapes = [
                    (keras.ops.shape(s)[2], keras.ops.shape(s)[3])
                    for s in sources_for_shape_derivation
                ]
            else:
                spatial_shapes = [
                    (keras.ops.shape(s)[1], keras.ops.shape(s)[2])
                    for s in sources_for_shape_derivation
                ]

        if spatial_shapes is None:
            spatial_shapes = [
                (
                    keras.ops.cast(self.anchor_image_size[0] / s, "int32"),
                    keras.ops.cast(self.anchor_image_size[1] / s, "int32"),
                )
                for s in self.feat_strides
            ]

        anchors = []
        for level, (height, width) in enumerate(spatial_shapes):
            grid_y, grid_x = keras.ops.meshgrid(
                keras.ops.arange(height, dtype=self.compute_dtype),
                keras.ops.arange(width, dtype=self.compute_dtype),
                indexing="ij",
            )
            grid_xy = keras.ops.stack([grid_x, grid_y], axis=-1)
            grid_xy = keras.ops.expand_dims(grid_xy, axis=0) + 0.5
            grid_xy = grid_xy / keras.ops.array(
                [width, height], dtype=self.compute_dtype
            )
            wh = keras.ops.ones_like(grid_xy) * grid_size * (2.0**level)
            level_anchors = keras.ops.concatenate([grid_xy, wh], axis=-1)
            level_anchors = keras.ops.reshape(
                level_anchors, (-1, height * width, 4)
            )
            anchors.append(level_anchors)

        eps = 1e-2
        anchors = keras.ops.concatenate(anchors, axis=1)
        valid_mask = keras.ops.all(
            (anchors > eps) & (anchors < 1 - eps), axis=-1, keepdims=True
        )
        anchors_transformed = keras.ops.log(anchors / (1 - anchors))
        dtype_name = keras.backend.standardize_dtype(self.compute_dtype)
        if dtype_name == "float16":
            finfo_dtype = np.float16
        else:
            finfo_dtype = np.float32
        max_float = keras.ops.array(
            np.finfo(finfo_dtype).max, dtype=self.compute_dtype
        )
        anchors = keras.ops.where(valid_mask, anchors_transformed, max_float)

        return anchors, valid_mask

    def compute_output_shape(
        self, sources_for_shape_derivation_shape=None, grid_size_shape=None
    ):
        num_total_anchors_dim = None

        if sources_for_shape_derivation_shape is None:
            num_total_anchors_calc = 0
            for s_stride in self.feat_strides:
                h = self.anchor_image_size[0] // s_stride
                w = self.anchor_image_size[1] // s_stride
                num_total_anchors_calc += h * w
            num_total_anchors_dim = num_total_anchors_calc
        else:
            calculated_spatial_elements = []
            for s_shape in sources_for_shape_derivation_shape:
                if self.data_format == "channels_first":
                    h, w = s_shape[2], s_shape[3]
                else:
                    h, w = s_shape[1], s_shape[2]
                if h is None or w is None:
                    calculated_spatial_elements.append(None)
                else:
                    calculated_spatial_elements.append(h * w)
            if any(elem is None for elem in calculated_spatial_elements):
                num_total_anchors_dim = None
            else:
                num_total_anchors_dim = sum(calculated_spatial_elements)

        anchors_shape = (1, num_total_anchors_dim, 4)
        valid_mask_shape = (1, num_total_anchors_dim, 1)
        return anchors_shape, valid_mask_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "anchor_image_size": self.anchor_image_size,
                "feat_strides": self.feat_strides,
                "data_format": self.data_format,
            }
        )
        return config


class DFineSpatialShapesExtractor(keras.layers.Layer):
    """Layer to extract spatial shapes from input tensors.

    This layer is used in `DFineBackbone` to extract the spatial dimensions
    (height, width) from the multi-scale feature maps. The resulting shape
    tensor is passed to the `DFineDecoder` for use in deformable attention.

    Args:
        data_format: str, optional, The data format of the input tensors.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(self, data_format=None, dtype=None, **kwargs):
        super().__init__(dtype=dtype, **kwargs)
        self.data_format = data_format

    def call(self, sources):
        if self.data_format == "channels_first":
            spatial_shapes = [
                (keras.ops.shape(s)[2], keras.ops.shape(s)[3]) for s in sources
            ]
        else:
            spatial_shapes = [
                (keras.ops.shape(s)[1], keras.ops.shape(s)[2]) for s in sources
            ]
        spatial_shapes_tensor = keras.ops.array(spatial_shapes, dtype="int32")
        return spatial_shapes_tensor

    def compute_output_shape(self, input_shape):
        if not isinstance(input_shape, list):
            raise ValueError("Expected a list of shape tuples")
        num_sources = len(input_shape)
        return (num_sources, 2)

    def get_config(self):
        config = super().get_config()
        config.update({"data_format": self.data_format})
        return config


class DFineInitialQueryAndReferenceGenerator(keras.layers.Layer):
    """Layer to generate initial queries and reference points for the decoder.

    This layer is a crucial component in `DFineBackbone` that bridges the
    encoder and decoder. It selects the top-k predictions from the encoder's
    output heads and uses them to generate the initial `target` (queries) and
    `reference_points` that are fed into the `DFineDecoder`.

    Args:
        num_queries: int, The number of queries to generate.
        hidden_dim: int, The hidden dimension of the model.
        learn_initial_query: bool, Whether to learn the initial query
            embeddings.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        num_queries,
        hidden_dim,
        learn_initial_query,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.num_queries = num_queries
        self.hidden_dim = hidden_dim
        self.learn_initial_query = learn_initial_query
        if self.learn_initial_query:
            self.query_indices_base = keras.ops.expand_dims(
                keras.ops.arange(self.num_queries, dtype="int32"), axis=0
            )
            self.weight_embedding = keras.layers.Embedding(
                input_dim=num_queries,
                output_dim=hidden_dim,
                name="weight_embedding",
                dtype=self.dtype_policy,
                embeddings_initializer="glorot_uniform",
            )
        else:
            self.weight_embedding = None

    def call(
        self,
        inputs,
        denoising_bbox_unact=None,
        denoising_class=None,
        training=None,
    ):
        (
            enc_outputs_class,
            enc_outputs_coord_logits_plus_anchors,
            output_memory,
            sources_last_element,
        ) = inputs
        enc_outputs_class_max = keras.ops.max(enc_outputs_class, axis=-1)
        topk_ind = keras.ops.top_k(
            enc_outputs_class_max, k=self.num_queries, sorted=True
        )[1]

        def gather_batch(elems):
            data, indices = elems
            return keras.ops.take(data, indices, axis=0)

        reference_points_unact = keras.ops.map(
            gather_batch, (enc_outputs_coord_logits_plus_anchors, topk_ind)
        )
        enc_topk_logits = keras.ops.map(
            gather_batch, (enc_outputs_class, topk_ind)
        )
        enc_topk_bboxes = keras.ops.sigmoid(reference_points_unact)

        if denoising_bbox_unact is not None:
            current_batch_size = keras.ops.shape(reference_points_unact)[0]
            denoising_bbox_unact = denoising_bbox_unact[:current_batch_size]
            if denoising_class is not None:
                denoising_class = denoising_class[:current_batch_size]
            reference_points_unact = keras.ops.concatenate(
                [denoising_bbox_unact, reference_points_unact], axis=1
            )
        if self.learn_initial_query:
            query_indices = self.query_indices_base
            target_embedding_val = self.weight_embedding(
                query_indices, training=training
            )
            batch_size = keras.ops.shape(sources_last_element)[0]
            target = keras.ops.tile(target_embedding_val, [batch_size, 1, 1])
        else:
            target = keras.ops.map(gather_batch, (output_memory, topk_ind))
            target = keras.ops.stop_gradient(target)

        if denoising_class is not None:
            target = keras.ops.concatenate([denoising_class, target], axis=1)
        init_reference_points = keras.ops.stop_gradient(reference_points_unact)
        return init_reference_points, target, enc_topk_logits, enc_topk_bboxes

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_queries": self.num_queries,
                "hidden_dim": self.hidden_dim,
                "learn_initial_query": self.learn_initial_query,
            }
        )
        return config

    def compute_output_spec(
        self,
        inputs,
        denoising_bbox_unact=None,
        denoising_class=None,
        training=None,
    ):
        (
            enc_outputs_class_spec,
            _,
            output_memory_spec,
            _,
        ) = inputs
        batch_size = enc_outputs_class_spec.shape[0]
        d_model_dim = output_memory_spec.shape[-1]
        num_labels_dim = enc_outputs_class_spec.shape[-1]
        num_queries_for_ref_points = self.num_queries
        if denoising_bbox_unact is not None:
            if len(denoising_bbox_unact.shape) > 1:
                if denoising_bbox_unact.shape[1] is not None:
                    num_queries_for_ref_points = (
                        denoising_bbox_unact.shape[1] + self.num_queries
                    )
                else:
                    num_queries_for_ref_points = None
        num_queries_for_target = self.num_queries
        if denoising_class is not None:
            if len(denoising_class.shape) > 1:
                if denoising_class.shape[1] is not None:
                    num_queries_for_target = (
                        denoising_class.shape[1] + self.num_queries
                    )
                else:
                    num_queries_for_target = None
        init_reference_points_spec = keras.KerasTensor(
            shape=(batch_size, num_queries_for_ref_points, 4),
            dtype=self.compute_dtype,
        )
        target_spec = keras.KerasTensor(
            shape=(batch_size, num_queries_for_target, d_model_dim),
            dtype=self.compute_dtype,
        )
        enc_topk_logits_spec = keras.KerasTensor(
            shape=(batch_size, self.num_queries, num_labels_dim),
            dtype=self.compute_dtype,
        )
        enc_topk_bboxes_spec = keras.KerasTensor(
            shape=(batch_size, self.num_queries, 4), dtype=self.compute_dtype
        )

        return (
            init_reference_points_spec,
            target_spec,
            enc_topk_logits_spec,
            enc_topk_bboxes_spec,
        )


class DFineIntegral(keras.layers.Layer):
    """Layer to compute integrated values from predicted corner probabilities.

    This layer implements the integral regression technique for bounding box
    prediction. It is used in `DFineDecoder` to transform the predicted
    distribution over bins (from `bbox_embed`) into continuous distance values,
    which are then used to calculate the final box coordinates.

    Args:
        max_num_bins: int, The maximum number of bins for the predictions.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(self, max_num_bins, dtype=None, **kwargs):
        super().__init__(dtype=dtype, **kwargs)
        self.max_num_bins = max_num_bins

    def build(self, input_shape):
        super().build(input_shape)

    def call(self, pred_corners, project, training=None):
        original_shape = keras.ops.shape(pred_corners)
        batch_size = original_shape[0]
        num_queries = original_shape[1]
        reshaped_pred_corners = keras.ops.reshape(
            pred_corners, (-1, self.max_num_bins + 1)
        )
        softmax_output = keras.ops.softmax(reshaped_pred_corners, axis=1)
        linear_output = keras.ops.matmul(
            softmax_output, keras.ops.transpose(project)
        )
        squeezed_output = keras.ops.squeeze(linear_output, axis=-1)
        output_grouped_by_4 = keras.ops.reshape(squeezed_output, (-1, 4))
        final_output = keras.ops.reshape(
            output_grouped_by_4, (batch_size, num_queries, -1)
        )
        return final_output

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "max_num_bins": self.max_num_bins,
            }
        )
        return config


class DFineLQE(keras.layers.Layer):
    """Layer to compute quality scores for predictions.

    This layer, used within `DFineDecoder`, implements the Localization Quality
    Estimation (LQE) head. It computes a quality score from the distribution of
    predicted bounding box corners and adds this score to the classification
    logits, enhancing prediction confidence.

    Args:
        top_prob_values: int, The number of top probabilities to consider.
        max_num_bins: int, The maximum number of bins for the predictions.
        lqe_hidden_dim: int, The hidden dimension for the MLP.
        num_lqe_layers: int, The number of layers in the MLP.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        top_prob_values,
        max_num_bins,
        lqe_hidden_dim,
        num_lqe_layers,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.top_prob_values = top_prob_values
        self.max_num_bins = max_num_bins
        self.reg_conf = DFineMLP(
            input_dim=4 * (self.top_prob_values + 1),
            hidden_dim=lqe_hidden_dim,
            output_dim=1,
            num_layers=num_lqe_layers,
            dtype=self.dtype_policy,
            last_layer_initializer="zeros",
            name="reg_conf",
        )

    def build(self, input_shape):
        reg_conf_input_shape = (
            input_shape[0][0],
            input_shape[0][1],
            4 * (self.top_prob_values + 1),
        )
        self.reg_conf.build(reg_conf_input_shape)
        super().build(input_shape)

    def call(self, scores, pred_corners, training=None):
        original_shape = keras.ops.shape(pred_corners)
        batch_size = original_shape[0]
        length = original_shape[1]
        reshaped_pred_corners = keras.ops.reshape(
            pred_corners, (batch_size, length, 4, self.max_num_bins + 1)
        )
        prob = keras.ops.softmax(reshaped_pred_corners, axis=-1)
        prob_topk, _ = keras.ops.top_k(
            prob, k=self.top_prob_values, sorted=True
        )
        stat = keras.ops.concatenate(
            [prob_topk, keras.ops.mean(prob_topk, axis=-1, keepdims=True)],
            axis=-1,
        )
        reshaped_stat = keras.ops.reshape(stat, (batch_size, length, -1))
        quality_score = self.reg_conf(reshaped_stat, training=training)
        return scores + quality_score

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "top_prob_values": self.top_prob_values,
                "max_num_bins": self.max_num_bins,
                "lqe_hidden_dim": self.reg_conf.hidden_dim,
                "num_lqe_layers": self.reg_conf.num_layers,
            }
        )
        return config


class DFineConvNormLayer(keras.layers.Layer):
    """Convolutional layer with normalization and optional activation.

    This is a fundamental building block used in the CNN parts of D-FINE. It
    combines a `Conv2D` layer with `BatchNormalization` and an optional
    activation. It is used extensively in layers like `DFineRepVggBlock`,
    `DFineCSPRepLayer`, and within the `DFineHybridEncoder`.

    Args:
        filters: int, The number of output channels.
        kernel_size: int, The size of the convolutional kernel.
        batch_norm_eps: float, The epsilon value for batch normalization.
        stride: int, The stride of the convolution.
        groups: int, The number of groups for grouped convolution.
        padding: int or None, The padding to apply.
        activation_function: str or None, The activation function to use.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
        channel_axis: int, optional, The channel axis. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        filters,
        kernel_size,
        batch_norm_eps,
        stride,
        groups,
        padding,
        activation_function,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.batch_norm_eps = batch_norm_eps
        self.stride = stride
        self.groups = groups
        self.padding_arg = padding
        self.activation_function = activation_function
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.channel_axis = channel_axis
        if self.padding_arg is None:
            keras_conv_padding_mode = "same"
            self.explicit_padding_layer = None
        else:
            keras_conv_padding_mode = "valid"
            self.explicit_padding_layer = keras.layers.ZeroPadding2D(
                padding=self.padding_arg,
                name=f"{self.name}_explicit_padding",
                dtype=self.dtype_policy,
            )

        self.convolution = keras.layers.Conv2D(
            filters=self.filters,
            kernel_size=self.kernel_size,
            strides=self.stride,
            padding=keras_conv_padding_mode,
            groups=self.groups,
            use_bias=False,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            name=f"{self.name}_convolution",
        )
        self.normalization = keras.layers.BatchNormalization(
            epsilon=self.batch_norm_eps,
            name=f"{self.name}_normalization",
            axis=self.channel_axis,
            dtype=self.dtype_policy,
        )
        self.activation_layer = (
            keras.layers.Activation(
                self.activation_function,
                name=f"{self.name}_activation",
                dtype=self.dtype_policy,
            )
            if self.activation_function
            else keras.layers.Identity(
                name=f"{self.name}_identity_activation", dtype=self.dtype_policy
            )
        )

    def build(self, input_shape):
        if self.explicit_padding_layer:
            self.explicit_padding_layer.build(input_shape)
            shape = self.explicit_padding_layer.compute_output_shape(
                input_shape
            )
        else:
            shape = input_shape
        self.convolution.build(shape)
        conv_output_shape = self.convolution.compute_output_shape(shape)
        self.normalization.build(conv_output_shape)
        self.activation_layer.build(conv_output_shape)
        super().build(input_shape)

    def call(self, hidden_state, training=None):
        if self.explicit_padding_layer:
            hidden_state = self.explicit_padding_layer(hidden_state)
        hidden_state = self.convolution(hidden_state)
        hidden_state = self.normalization(hidden_state, training=training)
        hidden_state = self.activation_layer(hidden_state)
        return hidden_state

    def compute_output_shape(self, input_shape):
        shape = input_shape
        if self.explicit_padding_layer:
            shape = self.explicit_padding_layer.compute_output_shape(shape)
        return self.convolution.compute_output_shape(shape)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "filters": self.filters,
                "kernel_size": self.kernel_size,
                "batch_norm_eps": self.batch_norm_eps,
                "stride": self.stride,
                "groups": self.groups,
                "padding": self.padding_arg,
                "activation_function": self.activation_function,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "channel_axis": self.channel_axis,
            }
        )
        return config


class DFineRepVggBlock(keras.layers.Layer):
    """RepVGG-style block with two parallel convolutional paths.

    This layer implements a block inspired by the RepVGG architecture, featuring
    two parallel convolutional paths (3x3 and 1x1) that are summed. It serves
    as the core bottleneck block within the `DFineCSPRepLayer`.

    Args:
        activation_function: str, The activation function to use.
        filters: int, The number of output channels.
        batch_norm_eps: float, The epsilon value for batch normalization.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
        channel_axis: int, optional, The channel axis. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        activation_function,
        filters,
        batch_norm_eps=1e-5,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.activation_function = activation_function
        self.filters = filters
        self.batch_norm_eps = batch_norm_eps
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.channel_axis = channel_axis
        self.conv1_layer = DFineConvNormLayer(
            filters=self.filters,
            kernel_size=3,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=1,
            activation_function=None,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv1",
        )
        self.conv2_layer = DFineConvNormLayer(
            filters=self.filters,
            kernel_size=1,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=0,
            activation_function=None,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv2",
        )
        self.activation_layer = (
            keras.layers.Activation(
                self.activation_function,
                name="block_activation",
                dtype=self.dtype_policy,
            )
            if self.activation_function
            else keras.layers.Identity(
                name="identity_activation", dtype=self.dtype_policy
            )
        )

    def build(self, input_shape):
        self.conv1_layer.build(input_shape)
        self.conv2_layer.build(input_shape)
        self.activation_layer.build(input_shape)
        super().build(input_shape)

    def call(self, x, training=None):
        y1 = self.conv1_layer(x, training=training)
        y2 = self.conv2_layer(x, training=training)
        y = y1 + y2
        return self.activation_layer(y)

    def compute_output_shape(self, input_shape):
        return self.conv1_layer.compute_output_shape(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "activation_function": self.activation_function,
                "filters": self.filters,
                "batch_norm_eps": self.batch_norm_eps,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "channel_axis": self.channel_axis,
            }
        )
        return config


class DFineCSPRepLayer(keras.layers.Layer):
    """CSP (Cross Stage Partial) layer with repeated bottleneck blocks.

    This layer implements a Cross Stage Partial (CSP) block using
    `DFineRepVggBlock` as its bottleneck. It is a key component of the
    `DFineFeatureAggregationBlock` block, which forms the FPN/PAN structure in
    the `DFineHybridEncoder`.

    Args:
        activation_function: str, The activation function to use.
        batch_norm_eps: float, The epsilon value for batch normalization.
        filters: int, The number of output channels.
        num_blocks: int, The number of bottleneck blocks.
        expansion: float, The expansion factor for hidden channels. Defaults to
            `1.0`.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
        channel_axis: int, optional, The channel axis. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        activation_function,
        batch_norm_eps,
        filters,
        num_blocks,
        expansion=1.0,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.activation_function = activation_function
        self.batch_norm_eps = batch_norm_eps
        self.filters = filters
        self.num_blocks = num_blocks
        self.expansion = expansion
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.channel_axis = channel_axis
        hidden_channels = int(self.filters * self.expansion)
        self.conv1 = DFineConvNormLayer(
            filters=hidden_channels,
            kernel_size=1,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=0,
            activation_function=self.activation_function,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv1",
        )
        self.conv2 = DFineConvNormLayer(
            filters=hidden_channels,
            kernel_size=1,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=0,
            activation_function=self.activation_function,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv2",
        )
        self.bottleneck_layers = [
            DFineRepVggBlock(
                activation_function=self.activation_function,
                filters=hidden_channels,
                batch_norm_eps=self.batch_norm_eps,
                dtype=self.dtype_policy,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                channel_axis=self.channel_axis,
                name=f"bottleneck_{i}",
            )
            for i in range(self.num_blocks)
        ]
        if hidden_channels != self.filters:
            self.conv3 = DFineConvNormLayer(
                filters=self.filters,
                kernel_size=1,
                batch_norm_eps=self.batch_norm_eps,
                stride=1,
                groups=1,
                padding=0,
                activation_function=self.activation_function,
                dtype=self.dtype_policy,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                channel_axis=self.channel_axis,
                name="conv3",
            )
        else:
            self.conv3 = keras.layers.Identity(
                name="conv3_identity", dtype=self.dtype_policy
            )

    def build(self, input_shape):
        self.conv1.build(input_shape)
        self.conv2.build(input_shape)
        bottleneck_input_shape = self.conv1.compute_output_shape(input_shape)
        for bottleneck_layer in self.bottleneck_layers:
            bottleneck_layer.build(bottleneck_input_shape)
        self.conv3.build(bottleneck_input_shape)
        super().build(input_shape)

    def call(self, hidden_state, training=None):
        hidden_state_1 = self.conv1(hidden_state, training=training)
        for bottleneck_layer in self.bottleneck_layers:
            hidden_state_1 = bottleneck_layer(hidden_state_1, training=training)
        hidden_state_2 = self.conv2(hidden_state, training=training)
        summed_hidden_states = hidden_state_1 + hidden_state_2
        if isinstance(self.conv3, keras.layers.Identity):
            hidden_state_3 = self.conv3(summed_hidden_states)
        else:
            hidden_state_3 = self.conv3(summed_hidden_states, training=training)
        return hidden_state_3

    def compute_output_shape(self, input_shape):
        shape_after_conv1 = self.conv1.compute_output_shape(input_shape)
        return self.conv3.compute_output_shape(shape_after_conv1)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "activation_function": self.activation_function,
                "batch_norm_eps": self.batch_norm_eps,
                "filters": self.filters,
                "num_blocks": self.num_blocks,
                "expansion": self.expansion,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "channel_axis": self.channel_axis,
            }
        )
        return config


class DFineFeatureAggregationBlock(keras.layers.Layer):
    """Complex block combining convolutional and CSP layers.

    This layer implements a complex feature extraction block combining multiple
    convolutional and `DFineCSPRepLayer` layers. It is the main building block
    for the Feature Pyramid Network (FPN) and Path Aggregation Network (PAN)
    pathways within the `DFineHybridEncoder`.

    Args:
        encoder_hidden_dim: int, The hidden dimension of the encoder.
        hidden_expansion: float, The expansion factor for hidden channels.
        batch_norm_eps: float, The epsilon value for batch normalization.
        activation_function: str, The activation function to use.
        num_blocks: int, The number of blocks in the CSP layers.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
        channel_axis: int, optional, The channel axis. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        encoder_hidden_dim,
        hidden_expansion,
        batch_norm_eps,
        activation_function,
        num_blocks,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.encoder_hidden_dim = encoder_hidden_dim
        self.hidden_expansion = hidden_expansion
        self.batch_norm_eps = batch_norm_eps
        self.activation_function = activation_function
        self.num_blocks = num_blocks
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.channel_axis = channel_axis

        conv3_dim = self.encoder_hidden_dim * 2
        self.conv4_dim = int(
            self.hidden_expansion * self.encoder_hidden_dim / 2
        )
        self.conv_dim = conv3_dim // 2
        self.conv1 = DFineConvNormLayer(
            filters=conv3_dim,
            kernel_size=1,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=0,
            activation_function=self.activation_function,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv1",
        )
        self.csp_rep1 = DFineCSPRepLayer(
            activation_function=self.activation_function,
            batch_norm_eps=self.batch_norm_eps,
            filters=self.conv4_dim,
            num_blocks=self.num_blocks,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="csp_rep1",
        )
        self.conv2 = DFineConvNormLayer(
            filters=self.conv4_dim,
            kernel_size=3,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=1,
            activation_function=self.activation_function,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv2",
        )
        self.csp_rep2 = DFineCSPRepLayer(
            activation_function=self.activation_function,
            batch_norm_eps=self.batch_norm_eps,
            filters=self.conv4_dim,
            num_blocks=self.num_blocks,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="csp_rep2",
        )
        self.conv3 = DFineConvNormLayer(
            filters=self.conv4_dim,
            kernel_size=3,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=1,
            activation_function=self.activation_function,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv3",
        )
        self.conv4 = DFineConvNormLayer(
            filters=self.encoder_hidden_dim,
            kernel_size=1,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=0,
            activation_function=self.activation_function,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv4",
        )

    def build(self, input_shape):
        self.conv1.build(input_shape)
        shape_after_conv1 = self.conv1.compute_output_shape(input_shape)
        csp_rep_input_shape_list = list(shape_after_conv1)
        csp_rep_input_shape_list[self.channel_axis] = self.conv_dim
        csp_rep_input_shape = tuple(csp_rep_input_shape_list)
        self.csp_rep1.build(csp_rep_input_shape)
        shape_after_csp_rep1 = self.csp_rep1.compute_output_shape(
            csp_rep_input_shape
        )
        self.conv2.build(shape_after_csp_rep1)
        shape_after_conv2 = self.conv2.compute_output_shape(
            shape_after_csp_rep1
        )
        self.csp_rep2.build(shape_after_conv2)
        shape_after_csp_rep2 = self.csp_rep2.compute_output_shape(
            shape_after_conv2
        )
        self.conv3.build(shape_after_csp_rep2)
        shape_for_concat_list = list(shape_after_conv1)
        shape_for_concat_list[self.channel_axis] = (
            self.conv_dim * 2 + self.conv4_dim * 2
        )
        shape_for_concat = tuple(shape_for_concat_list)
        self.conv4.build(shape_for_concat)
        super().build(input_shape)

    def call(self, input_features, training=None):
        conv1_out = self.conv1(input_features, training=training)
        split_features_tensor = keras.ops.split(
            conv1_out, [self.conv_dim, self.conv_dim], axis=self.channel_axis
        )
        split_features = list(split_features_tensor)
        branch1 = self.csp_rep1(split_features[-1], training=training)
        branch1 = self.conv2(branch1, training=training)
        branch2 = self.csp_rep2(branch1, training=training)
        branch2 = self.conv3(branch2, training=training)
        split_features.extend([branch1, branch2])
        merged_features = keras.ops.concatenate(
            split_features, axis=self.channel_axis
        )
        merged_features = self.conv4(merged_features, training=training)
        return merged_features

    def compute_output_shape(self, input_shape):
        shape_after_conv1 = self.conv1.compute_output_shape(input_shape)
        shape_for_concat_list = list(shape_after_conv1)
        shape_for_concat_list[self.channel_axis] = (
            self.conv_dim * 2 + self.conv4_dim * 2
        )
        shape_for_concat = tuple(shape_for_concat_list)
        return self.conv4.compute_output_shape(shape_for_concat)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "encoder_hidden_dim": self.encoder_hidden_dim,
                "hidden_expansion": self.hidden_expansion,
                "batch_norm_eps": self.batch_norm_eps,
                "activation_function": self.activation_function,
                "num_blocks": self.num_blocks,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "channel_axis": self.channel_axis,
            }
        )
        return config


class DFineSCDown(keras.layers.Layer):
    """Downsampling layer using convolutions.

    This layer is used in the `DFineHybridEncoder` to perform downsampling.
    Specifically, it is part of the Path Aggregation Network (PAN) bottom-up
    pathway, reducing the spatial resolution of feature maps.

    Args:
        encoder_hidden_dim: int, The hidden dimension of the encoder.
        batch_norm_eps: float, The epsilon value for batch normalization.
        kernel_size: int, The kernel size for the second convolution.
        stride: int, The stride for the second convolution.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
        channel_axis: int, optional, The channel axis. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        encoder_hidden_dim,
        batch_norm_eps,
        kernel_size,
        stride,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.encoder_hidden_dim = encoder_hidden_dim
        self.batch_norm_eps = batch_norm_eps
        self.conv2_kernel_size = kernel_size
        self.conv2_stride = stride
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.channel_axis = channel_axis
        self.conv1 = DFineConvNormLayer(
            filters=self.encoder_hidden_dim,
            kernel_size=1,
            batch_norm_eps=self.batch_norm_eps,
            stride=1,
            groups=1,
            padding=0,
            activation_function=None,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv1",
        )
        self.conv2 = DFineConvNormLayer(
            filters=self.encoder_hidden_dim,
            kernel_size=self.conv2_kernel_size,
            batch_norm_eps=self.batch_norm_eps,
            stride=self.conv2_stride,
            groups=self.encoder_hidden_dim,
            padding=(self.conv2_kernel_size - 1) // 2,
            activation_function=None,
            dtype=self.dtype_policy,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            channel_axis=self.channel_axis,
            name="conv2",
        )

    def build(self, input_shape):
        self.conv1.build(input_shape)
        shape_after_conv1 = self.conv1.compute_output_shape(input_shape)
        self.conv2.build(shape_after_conv1)
        super().build(input_shape)

    def call(self, input_features, training=None):
        x = self.conv1(input_features, training=training)
        x = self.conv2(x, training=training)
        return x

    def compute_output_shape(self, input_shape):
        shape_after_conv1 = self.conv1.compute_output_shape(input_shape)
        return self.conv2.compute_output_shape(shape_after_conv1)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "encoder_hidden_dim": self.encoder_hidden_dim,
                "batch_norm_eps": self.batch_norm_eps,
                "kernel_size": self.conv2_kernel_size,
                "stride": self.conv2_stride,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "channel_axis": self.channel_axis,
            }
        )
        return config


class DFineMLPPredictionHead(keras.layers.Layer):
    """MLP head for making predictions from feature vectors.

    This layer is a generic MLP used for various prediction tasks in D-FINE.
    It is used for the encoder's bounding box head (`enc_bbox_head` in
    `DFineBackbone`), the decoder's bounding box embedding (`bbox_embed` in
    `DFineDecoder`), and the query position head (`query_pos_head` in
    `DFineDecoder`).

    Args:
        input_dim: int, The input dimension.
        hidden_dim: int, The hidden dimension for intermediate layers.
        output_dim: int, The output dimension.
        num_layers: int, The number of layers in the MLP.
        kernel_initializer: str or Initializer, optional, Initializer for
            the kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or Initializer, optional, Initializer for
            the bias weights. Defaults to `"zeros"`.
        last_layer_initializer: str or Initializer, optional, Special
            initializer for the final layer's weights and biases. If `None`,
            uses `kernel_initializer` and `bias_initializer`. Defaults to
            `None`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        input_dim,
        hidden_dim,
        output_dim,
        num_layers,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        last_layer_initializer=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.num_layers = num_layers
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.last_layer_initializer = keras.initializers.get(
            last_layer_initializer
        )

        h = [self.hidden_dim] * (self.num_layers - 1)
        input_dims = [self.input_dim] + h
        output_dims = h + [self.output_dim]

        self.dense_layers = []
        for i, (_, out_dim) in enumerate(zip(input_dims, output_dims)):
            is_last_layer = i == self.num_layers - 1
            current_kernel_init = self.kernel_initializer
            current_bias_init = self.bias_initializer
            if is_last_layer and self.last_layer_initializer is not None:
                current_kernel_init = self.last_layer_initializer
                current_bias_init = self.last_layer_initializer
            self.dense_layers.append(
                keras.layers.Dense(
                    units=out_dim,
                    name=f"linear_{i}",
                    dtype=self.dtype_policy,
                    kernel_initializer=current_kernel_init,
                    bias_initializer=current_bias_init,
                )
            )

    def build(self, input_shape):
        if self.dense_layers:
            current_build_shape = input_shape
            for i, dense_layer in enumerate(self.dense_layers):
                dense_layer.build(current_build_shape)
                current_build_shape = dense_layer.compute_output_shape(
                    current_build_shape
                )
        super().build(input_shape)

    def call(self, x, training=None):
        current_x = x
        for i, layer in enumerate(self.dense_layers):
            current_x = layer(current_x)
            if i < self.num_layers - 1:
                current_x = keras.ops.relu(current_x)
        return current_x

    def compute_output_spec(self, x_spec):
        output_shape = list(x_spec.shape)
        output_shape[-1] = self.output_dim
        return keras.KerasTensor(
            shape=tuple(output_shape), dtype=self.compute_dtype
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_dim": self.input_dim,
                "hidden_dim": self.hidden_dim,
                "output_dim": self.output_dim,
                "num_layers": self.num_layers,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
                "last_layer_initializer": keras.initializers.serialize(
                    self.last_layer_initializer
                ),
            }
        )
        return config



import keras

try:
    import tensorflow as tf
except ImportError:
    tf = None

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.audio_converter import AudioConverter
from keras_hub.src.models.moonshine.moonshine_backbone import MoonshineBackbone


@keras_hub_export("keras_hub.layers.MoonshineAudioConverter")
class MoonshineAudioConverter(AudioConverter):
    """Moonshine audio preprocessing layer.

    This layer processes raw audio waveforms for the Moonshine ASR model. Audio
    is formatted as a batched tensor at a 16kHz sample rate and validated for
    length (0.1 to 64 seconds). The layer handles padding and optional
    normalization. It does not contain trainable weights.

    Args:
        sampling_rate: int, optional. The audio sampling rate in Hz. Defaults to
            16,000.
        padding_value: float, optional. The value for padding. Defaults to 0.0.
        do_normalize: bool, optional. Whether to normalize inputs. Defaults to
            False.
        **kwargs: Additional keyword arguments passed to the base AudioConverter
            class for customizing the underlying preprocessing behavior.

    Call arguments:
        - `inputs`: The raw audio data to be processed. It should be a tensor of
          shape `(batch_size, time_steps, 1)` for mono audio. If the input has
          shape `(batch_size, time_steps)`, the layer will add the channel
          dimension.
        - `sampling_rate`: The sampling rate of the audio in Hz. If
          provided, it must match the expected sampling rate set during
          initialization (default is 16,000 Hz). If not provided, the expected
          sampling rate is taken from the initialization arguments.
        - `padding`: The padding strategy to apply. If provided, can be one of:
            - `"longest"`: If `pad_to_multiple_of` is set, pads the audio to
              make the time_steps dimension a multiple of `pad_to_multiple_of`.
            - `"max_length"`: Pads or truncates the audio to `max_length` time
              steps. If `pad_to_multiple_of` is set, the target length will be
              the smallest multiple of `pad_to_multiple_of` that is greater than
              or equal to `max_length`.
            - If not specified or `None`, no padding is applied.
        - `max_length`: The target number of time steps when `padding` is
          `"max_length"`. If not provided and `padding` is `"max_length"`, no
          padding or truncation is applied.
        - `pad_to_multiple_of`: If set, the padded time_steps will be a
          multiple of this value for the chosen padding strategy.

    Examples:
    ```python
    import keras
    from keras_hub.layers import MoonshineAudioConverter

    # Create a dummy audio input (1 second at 16kHz).
    dummy_audio = keras.ops.convert_to_tensor(
        [[0.1] * 16000],
        dtype="float32"
    )
    dummy_audio = keras.ops.expand_dims(dummy_audio, axis=-1)

    # Initialize the preprocessor.
    preprocessor = MoonshineAudioConverter(do_normalize=True)

    # Process the audio.
    processed_audio = preprocessor(dummy_audio)

    # Output shape.
    print(processed_audio.shape) # Expected: (1, 16000, 1) or padded length
    ```
    """

    # References:
    # Defined and formulated based on the UsefulSensors implementation of audio
    # preprocessing logic (https://github.com/usefulsensors/moonshine/blob/main/moonshine/transcribe.py).

    backbone_cls = MoonshineBackbone

    def __init__(
        self,
        sampling_rate=16000,
        padding_value=0.0,
        do_normalize=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._convert_input_args = False
        self._allow_non_tensor_positional_args = True
        self.sampling_rate = sampling_rate
        self.padding_value = padding_value
        self.do_normalize = do_normalize

    def call(
        self,
        inputs,
        sampling_rate=None,
        padding=None,
        max_length=None,
        pad_to_multiple_of=None,
    ):
        # Validate sampling rate.
        if sampling_rate is not None and sampling_rate != self.sampling_rate:
            raise ValueError(
                f"Expected sampling_rate {self.sampling_rate}, got "
                f"{sampling_rate}"
            )

        # Ensure inputs are (batch_size, time_steps, 1).
        input_shape = keras.ops.shape(inputs)
        input_rank = len(input_shape)
        if input_rank == 2:
            processed_inputs = keras.ops.expand_dims(inputs, axis=-1)
        elif input_rank == 3:
            processed_inputs = inputs
        else:
            raise ValueError(
                "Inputs must be mono audio: (batch_size, time_steps, 1)"
            )

        # Get original length and validate duration.
        current_shape = keras.ops.shape(processed_inputs)
        original_length = current_shape[1]
        duration = (
            keras.ops.cast(original_length, keras.backend.floatx())
            / self.sampling_rate
        )
        # Source: https://github.com/usefulsensors/moonshine/blob/4a000427bd36a1c2c6d20a86c672dbd850b44c88/moonshine/transcribe.py#L20
        is_invalid_duration = keras.ops.logical_or(
            keras.ops.less(duration, 0.1), keras.ops.greater(duration, 64.0)
        )

        def print_warning_fn():
            import warnings

            warnings.warn(
                "Audio duration must be between 0.1 and 64 seconds. For "
                "transcribing longer segments, pre-segment your audio and "
                "provide shorter segments."
            )
            return keras.ops.convert_to_tensor(True, dtype="bool")

        is_tf_symbolic = (
            tf is not None
            and hasattr(processed_inputs, "graph")
            and hasattr(processed_inputs.graph, "as_graph_def")
        )
        use_tf_graph_ops = tf is not None and is_tf_symbolic
        if use_tf_graph_ops and keras.config.backend() != "torch":
            _ = tf.cond(
                is_invalid_duration,
                print_warning_fn,
                lambda: keras.ops.convert_to_tensor(False, dtype="bool"),
            )
        else:
            if keras.ops.convert_to_numpy(is_invalid_duration):
                print_warning_fn()

        # Handle padding.
        if padding == "longest":
            target_length = original_length
            if pad_to_multiple_of:
                target_length = (
                    (target_length + pad_to_multiple_of - 1)
                    // pad_to_multiple_of
                ) * pad_to_multiple_of

            needs_padding = keras.ops.greater(target_length, original_length)

            def pad_fn():
                padding_amount = target_length - original_length
                paddings = [[0, 0], [0, padding_amount], [0, 0]]
                if use_tf_graph_ops and keras.config.backend() != "tensorflow":
                    return tf.pad(
                        processed_inputs,
                        paddings,
                        mode="CONSTANT",
                        constant_values=float(self.padding_value),
                    )
                else:
                    return keras.ops.pad(
                        processed_inputs,
                        paddings,
                        mode="constant",
                        constant_values=self.padding_value,
                    )

            if use_tf_graph_ops and keras.config.backend() != "torch":
                processed_inputs = tf.cond(
                    needs_padding, pad_fn, lambda: processed_inputs
                )
            else:
                processed_inputs = keras.ops.cond(
                    needs_padding, pad_fn, lambda: processed_inputs
                )

        elif padding == "max_length" and max_length is not None:
            target_length_const = max_length
            if pad_to_multiple_of:
                target_length_const = (
                    (target_length_const + pad_to_multiple_of - 1)
                    // pad_to_multiple_of
                ) * pad_to_multiple_of

            needs_padding = keras.ops.less(original_length, target_length_const)
            needs_truncating = keras.ops.greater(
                original_length, target_length_const
            )

            def pad_fn():
                padding_amount = target_length_const - original_length
                paddings = [[0, 0], [0, padding_amount], [0, 0]]
                if use_tf_graph_ops and keras.config.backend() != "tensorflow":
                    return tf.pad(
                        processed_inputs,
                        paddings,
                        mode="CONSTANT",
                        constant_values=float(self.padding_value),
                    )
                else:
                    return keras.ops.pad(
                        processed_inputs,
                        paddings,
                        mode="constant",
                        constant_values=self.padding_value,
                    )

            def trunc_fn():
                if use_tf_graph_ops and keras.config.backend() != "tensorflow":
                    return processed_inputs[:, :target_length_const, :]
                else:
                    return keras.ops.slice(
                        processed_inputs,
                        [0, 0, 0],
                        [-1, target_length_const, -1],
                    )

            if use_tf_graph_ops and keras.config.backend() != "torch":
                processed_inputs = tf.cond(
                    needs_padding,
                    pad_fn,
                    lambda: tf.cond(
                        needs_truncating, trunc_fn, lambda: processed_inputs
                    ),
                )
            else:
                needs_padding = keras.ops.less(
                    original_length, target_length_const
                )
                needs_truncating = keras.ops.greater(
                    original_length, target_length_const
                )
                needs_padding_bool = keras.ops.convert_to_numpy(needs_padding)
                needs_truncating_bool = keras.ops.convert_to_numpy(
                    needs_truncating
                )

                if needs_padding_bool:
                    padding_amount = target_length_const - original_length
                    paddings = [[0, 0], [0, padding_amount], [0, 0]]
                    processed_inputs = keras.ops.pad(
                        processed_inputs,
                        paddings,
                        mode="constant",
                        constant_values=self.padding_value,
                    )
                elif needs_truncating_bool:
                    processed_inputs = processed_inputs[
                        :, :target_length_const, :
                    ]

        # Normalize if enabled.
        if self.do_normalize:
            mean = keras.ops.mean(processed_inputs, axis=1, keepdims=True)
            var = keras.ops.var(processed_inputs, axis=1, keepdims=True)
            processed_inputs = (processed_inputs - mean) / keras.ops.sqrt(
                var + 1e-7
            )

        return processed_inputs

    def compute_output_shape(self, input_shape):
        # [batch_size, time_steps] → [batch_size, time_steps, 1].
        if len(input_shape) == 2 or len(input_shape) == 3:
            return (input_shape[0], None, 1)
        else:
            raise ValueError("Input shape must be rank 2 or 3.")

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "sampling_rate": self.sampling_rate,
                "padding_value": self.padding_value,
                "do_normalize": self.do_normalize,
            }
        )
        return config
